/**
 * @file
 *
 * @author Lawrence Murray <lawrence.murray@csiro.au>
 */
#ifndef BI_CUDA_UPDATER_DYNAMICLOGDENSITYKERNEL_CUH
#define BI_CUDA_UPDATER_DYNAMICLOGDENSITYKERNEL_CUH

#include "../cuda.hpp"

namespace bi {
/**
 * Kernel function for dynamic log-density evaluation.
 *
 * @tparam B Model type.
 * @tparam S Action type list.
 * @tparam T1 Scalar type.
 * @tparam V1 Vector type.
 *
 * @param t1 Start time.
 * @param t2 End time.
 * @param[in,out] s State.
 * @param[in,out] lp Log-density.
 */
template<class B, class S, class T1, class V1>
CUDA_FUNC_GLOBAL void kernelDynamicLogDensity(const T1 t1, const T1 t2,
    State<B,ON_DEVICE> s, V1 lp);

}

#include "DynamicLogDensityMatrixVisitorGPU.cuh"
#include "DynamicLogDensityVisitorGPU.cuh"
#include "../constant.cuh"
#include "../global.cuh"
#include "../../state/Pa.hpp"

#include "boost/mpl/if.hpp"

template<class B, class S, class T1, class V1>
CUDA_FUNC_GLOBAL void bi::kernelDynamicLogDensity(const T1 t1, const T1 t2,
    State<B,ON_DEVICE> s, V1 lp) {
  typedef Pa<ON_DEVICE,B,global,global,global,global> PX;
  typedef Ou<ON_DEVICE,B,global> OX;
  typedef DynamicLogDensityMatrixVisitorGPU<B,S,T1,PX,OX> MatrixVisitor;
  typedef DynamicLogDensityVisitorGPU<B,S,T1,PX,OX> ElementVisitor;
  typedef typename boost::mpl::if_c<block_is_matrix<S>::value,MatrixVisitor,
      ElementVisitor>::type Visitor;

  const int p = blockIdx.x*blockDim.x + threadIdx.x;
  const int i = blockIdx.y*blockDim.y + threadIdx.y;
  PX pax;
  OX x;

  /* update */
  if (p < s.size()) {
    Visitor::accept(t1, t2, s, p, i, pax, x);
  }
}

#endif
